# 九、关系数据库和 SQL

> 原文：[Relational Databases and SQL](https://www.textbook.ds100.org/ch/09/sql_intro.html)
> 
> 校验：[Kitty Du](https://github.com/miaoxiaozui2017)

```py
# HIDDEN
# Clear previously defined variables
%reset -f

# Set directory for data loading to work properly
import os
os.chdir(os.path.expanduser('~/notebooks/09'))
```

到目前为止，我们已经处理了作为文本文件存储在计算机上的数据集。虽然使用文本文件存储数据对于分析小数据集很有用，但对许多实际用例来说却是一个挑战。

许多数据集是由多个人收集的，例如，一个数据科学家团队。但是，如果数据存储在文本文件中，那么团队可能需要在每次更新数据时发送和下载新版本的文件。文本文件本身并不能为多个分析员提供一致的数据检索点。这个问题，以及其他问题，使得文本文件难以用于更大的数据集或团队。

我们经常使用关系数据库管理系统（RDBMS）来存储数据，如 MySQL 或 PostgreSQL。为了使用这些系统，我们使用一种称为 SQL 的查询语言，而不是 Python。在本章中，我们将讨论关系数据库模型并介绍 SQL。

## 9.1 关系数据库

```py
# HIDDEN
# Clear previously defined variables
%reset -f

# Set directory for data loading to work properly
import os
os.chdir(os.path.expanduser('~/notebooks/09'))
```

**数据库**是有组织的数据集合。过去，数据存储在专门为特定任务设计的数据结构中。例如，航空公司记录航班预定的格式可能不同于银行管理账户分类账的格式。1969 年，Ted Codd 将关系模型作为一种存储数据的通用方法引入。数据存储在被称为**关系(relations)** 的二维表中，由每行中的单个观察值（通常称为**元组(tuples)** ）组成。每个元组都是一个结构化数据项，表示某些**属性(attributes)**(列) 之间的关系。关系的每个属性都有一个名称和数据类型。

考虑下面的`purchases`关系：

<header style="text-align:center">**purchases**</header>

| **name** | **product** | **retailer** | **data purchased** |
| --: | --: | --: | --:|
| Samantha | iPod | Best Buy | June 3,2016 |
| Timothy | Chromebook | Amazon | July 8,2016 |
| Jason | Surface Pro | Target | October 2,2016 |

在`purchases`中，每个元组表示`name`、`product`、`retailer`和`date purchased`属性之间的关系。

关系的 *模式(schema)* 包含其列名、数据类型和约束。例如，`purchases`表的模式说明列是`name`、`product`、`retailer`和`date purchased`；它还说明每个列都包含文本。

以下`prices`关系显示了一些零售商店中某些小物件的价格：

<header style="text-align:center">**prices**</header>

| **retailer** | **product** | **price** |
| --: | --: | --: |
| Best Buy | Galaxy S9 | 719.00 |
| Best Buy | iPod | 200.00 |
| Amazon | iPad | 450.00 |
| Amazon | Battery pack | 24.87 |
| Amazon | Chromebook | 249.99 |
| Target | iPod | 215.00 |
| Target | Surface Pro | 799.00 |
| Target | Google Pixel 2 | 659.00 |
| Walmart | Chromebook | 238.79 |

然后，我们可以同时参考这两个表来确定 Samantha、Timothy 和 Jason 为各自的小物件支付了多少钱（假设每家店的价格随时间保持不变）。这两个表一起构成了一个**关系数据库(relational database)** ，它是一个或多个关系的集合。整个数据库的模式是数据库中各个关系的模式集。

## 关系数据库管理系统[](#Relational-Database-Management-Systems)

关系数据库可以简单地描述为一组包含单个数据项行的表。关系数据库管理系统（RDBMSs）提供了到关系数据库的接口。[Oracle](https://www.wikiwand.com/en/Oracle_Database)、[MySQL](https://www.wikiwand.com/en/MySQL)和[PostgreSQL](https://www.wikiwand.com/en/PostgreSQL)是目前实践中最常用的三种 RDBMS。

关系数据库管理系统使用户能够添加、编辑和删除数据库中的数据。与使用文本文件集合存储数据相比，这些系统提供了几个关键好处，包括：

1.  可靠的数据存储：RDBMS 保护数据不受系统故障或崩溃的影响。
2.  性能：RDBMS 通常比文本文件更高效地存储数据，并且拥有完善的查询数据的算法。
3.  数据管理：RDBMSS 实现访问控制，防止未经授权的用户访问敏感数据集。
4.  数据一致性：RDBMSS 可以对输入的数据施加约束，例如，列`GPA`只包含介于 0.0 和 4.0 之间的浮点数。

为了处理存储在 RDBMS 中的数据，我们使用 SQL 编程语言。

## RDBMS 与 pandas 的对比[](#RDBMS-vs-pandas)

RDBMS 和 `pandas` Python 包有何区别？首先，`pandas`不关心数据存储。虽然 数据帧(DataFrame) 可以从多种数据格式中读写，但 `pandas` 并不像 RDBMS 那样规定数据实际存储在底层计算机上的方式。其次，`pandas` 主要提供操作数据的方法，而 RDBMS 同时处理数据存储和数据操作，使它们更适合于较大的数据集。一个典型的经验法则是对大于几 GB 的数据集使用 RDBMS。最后，`pandas` 需要了解 Python 才能使用，而 RDBMS 需要了解 SQL。由于 SQL 比 Python 学习起来简单，RDBMS 允许用户使用较少的技术就能存储和查询数据，这是一个方便的特性。

## 9.2 SQL

```py
# HIDDEN
# Clear previously defined variables
%reset -f

# Set directory for data loading to work properly
import os
os.chdir(os.path.expanduser('~/notebooks/09'))
```

```py
# HIDDEN
import warnings
# Ignore numpy dtype warnings. These warnings are caused by an interaction
# between numpy and Cython and can be safely ignored.
# Reference: https://stackoverflow.com/a/40846742
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
%matplotlib inline
import ipywidgets as widgets
from ipywidgets import interact, interactive, fixed, interact_manual
import nbinteract as nbi

sns.set()
sns.set_context('talk')
np.set_printoptions(threshold=20, precision=2, suppress=True)
pd.options.display.max_rows = 7
pd.options.display.max_columns = 8
pd.set_option('precision', 2)
# This option stops scientific notation for pandas
# pd.set_option('display.float_format', '{:.2f}'.format)
```

```py
# HIDDEN

# Creating a table
sql_expr = """
CREATE TABLE prices(
    retailer TEXT,
    product TEXT,
    price FLOAT);
"""
result = sqlite_engine.execute(sql_expr)
```

```py
# HIDDEN

# Inserting records into the table
sql_expr = """
INSERT INTO prices VALUES 
  ('Best Buy', 'Galaxy S9', 719.00),
  ('Best Buy', 'iPod', 200.00),
  ('Amazon', 'iPad', 450.00),
  ('Amazon', 'Battery pack',  24.87),
  ('Amazon', 'Chromebook', 249.99),
  ('Target', 'iPod', 215.00),
  ('Target', 'Surface Pro', 799.00),
  ('Target', 'Google Pixel 2', 659.00),
  ('Walmart', 'Chromebook', 238.79);
"""
result = sqlite_engine.execute(sql_expr)
```

```py
# HIDDEN
import pandas as pd

prices = pd.DataFrame([['Best Buy', 'Galaxy S9', 719.00],
                   ['Best Buy', 'iPod', 200.00],
                   ['Amazon', 'iPad', 450.00],
                   ['Amazon', 'Battery pack', 24.87],
                   ['Amazon', 'Chromebook', 249.99],
                   ['Target', 'iPod', 215.00],
                   ['Target', 'Surface Pro', 799.00],
                   ['Target', 'Google Pixel 2', 659.00],
                   ['Walmart', 'Chromebook', 238.79]],
                 columns=['retailer', 'product', 'price'])
```

**sql**（结构化查询语言）是一种编程语言，它具有对关系数据库管理系统（RDBMS）中存储的数据进行定义、逻辑组织、操作和执行计算的操作。

SQL 是一种声明性语言。这意味着用户只需要指定他们想要的 _ 数据类型，而不需要指定 _ 如何获取它。下面是一个示例，其中一个必要的示例用于比较：__

*   **声明性**：从表“a”计算列“x”和“y”，其中“y”中的值大于 100.00。
*   **命令式**：对于表“a”中的每个记录，检查记录是否包含大于 100 的“y”值。如果是，则将记录的“x”和“y”属性存储在新表中。返回新表。

在本章中，我们将以 python 字符串的形式编写 SQL 查询，然后使用`pandas`执行 SQL 查询并将结果读取到`pandas`数据帧中。在我们介绍 SQL 语法的基础知识时，为了便于比较，我们偶尔还会显示`pandas`等价物。

### 通过`pandas`[¶](#Executing-SQL-Queries-through-pandas)执行 SQL 查询

要从 python 执行 SQL 查询，我们将使用[sqlachemy](http://docs.sqlalchemy.org/en/latest/core/tutorial.html)库连接到数据库。然后我们可以使用`pandas`函数[pd.read_sql](https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_sql.html)通过这个连接执行 SQL 查询。

```py
import sqlalchemy

# pd.read_sql takes in a parameter for a SQLite engine, which we create below
sqlite_uri = "sqlite:///sql_basics.db"
sqlite_engine = sqlalchemy.create_engine(sqlite_uri)
```

此数据库包含一个关系：`prices`。为了显示关系，我们运行一个 SQL 查询。调用`read_sql`将在 RDBMS 上执行 SQL 查询，然后在`pandas`数据帧中返回结果。

```py
sql_expr = """
SELECT * 
FROM prices
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | 零售商 | 产品 | 价格 |
| --- | --- | --- | --- |
| 零 | 百思买 | 银河 S9 | 七百一十九 |
| --- | --- | --- | --- |
| 1 个 | Best Buy | iPod | 二百 |
| --- | --- | --- | --- |
| 二 | 亚马逊 | iPad | 四百五十 |
| --- | --- | --- | --- |
| 三 | Amazon | 电池组 | 二十四点八七 |
| --- | --- | --- | --- |
| 四 | Amazon | 彩色电子书 | 二百四十九点九九 |
| --- | --- | --- | --- |
| 5 个 | 目标 | iPod | 二百一十五 |
| --- | --- | --- | --- |
| 六 | Target | 微软 Surface Pro 平板电脑 | 七百九十九 |
| --- | --- | --- | --- |
| 七 | Target | 谷歌像素 2 | 六百五十九 |
| --- | --- | --- | --- |
| 8 个 | 沃尔玛 | Chromebook | 二百三十八点七九 |
| --- | --- | --- | --- |

在本节后面，我们将比较 SQL 查询和`pandas`方法调用，因此我们在`pandas`中创建了相同的数据帧。

```py
prices
```

|  | retailer | product | price |
| --- | --- | --- | --- |
| 0 | Best Buy | Galaxy S9 | 719.00 |
| --- | --- | --- | --- |
| 1 | Best Buy | iPod | 200.00 |
| --- | --- | --- | --- |
| 2 | Amazon | iPad | 450.00 |
| --- | --- | --- | --- |
| 3 | Amazon | Battery pack | 24.87 |
| --- | --- | --- | --- |
| 4 | Amazon | Chromebook | 249.99 |
| --- | --- | --- | --- |
| 5 | Target | iPod | 215.00 |
| --- | --- | --- | --- |
| 6 | Target | Surface Pro | 799.00 |
| --- | --- | --- | --- |
| 7 | Target | Google Pixel 2 | 659.00 |
| --- | --- | --- | --- |
| 8 | Walmart | Chromebook | 238.79 |
| --- | --- | --- | --- |

## SQL 句法[¶](#SQL-Syntax)

所有 SQL 查询都采用以下常规形式：

```py
SELECT [DISTINCT] <column expression list>
FROM <relation>
[WHERE <predicate>]
[GROUP BY <column list>]
[HAVING <predicate>]
[ORDER BY <column list>]
[LIMIT <number>]
```

注意：

1.  **方括号中的所有内容都是可选的。**有效的 SQL 查询只需要`SELECT`和`FROM`语句。
2.  **SQL 语法通常用大写字母编写。**虽然不需要大写，但通常使用大写字母编写 SQL 语法。它还可以帮助您直观地构造查询，以供其他人阅读。
3.  `FROM`查询块可以引用一个或多个表，尽管在本节中，为了简单起见，我们一次只查看一个表。

### 从[¶](#SELECT-and-FROM)中选择和

SQL 查询中的两个强制语句是：

*   `SELECT`指示要查看的列。
*   `FROM`指示从中选择这些列的表。

要显示整个`prices`表，我们运行：

```py
sql_expr = """
SELECT * 
FROM prices
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer | product | price |
| --- | --- | --- | --- |
| 0 | Best Buy | Galaxy S9 | 719.00 |
| --- | --- | --- | --- |
| 1 | Best Buy | iPod | 200.00 |
| --- | --- | --- | --- |
| 2 | Amazon | iPad | 450.00 |
| --- | --- | --- | --- |
| 3 | Amazon | Battery pack | 24.87 |
| --- | --- | --- | --- |
| 4 | Amazon | Chromebook | 249.99 |
| --- | --- | --- | --- |
| 5 | Target | iPod | 215.00 |
| --- | --- | --- | --- |
| 6 | Target | Surface Pro | 799.00 |
| --- | --- | --- | --- |
| 7 | Target | Google Pixel 2 | 659.00 |
| --- | --- | --- | --- |
| 8 | Walmart | Chromebook | 238.79 |
| --- | --- | --- | --- |

`SELECT *`返回原始关系中的每一列。为了只显示`prices`中表示的零售商，我们将`retailer`列添加到`SELECT`语句中。

```py
sql_expr = """
SELECT retailer
FROM prices
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer |
| --- | --- |
| 0 | Best Buy |
| --- | --- |
| 1 | Best Buy |
| --- | --- |
| 2 | Amazon |
| --- | --- |
| 3 | Amazon |
| --- | --- |
| 4 | Amazon |
| --- | --- |
| 5 | Target |
| --- | --- |
| 6 | Target |
| --- | --- |
| 7 | Target |
| --- | --- |
| 8 | Walmart |
| --- | --- |

如果我们想要一个唯一零售商的列表，我们可以调用`DISTINCT`函数来省略重复的值。

```py
sql_expr = """
SELECT DISTINCT(retailer)
FROM prices
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer |
| --- | --- |
| 0 | Best Buy |
| --- | --- |
| 1 | Amazon |
| --- | --- |
| 2 | Target |
| --- | --- |
| 3 | Walmart |
| --- | --- |

这相当于以下`pandas`代码的功能：

```py
prices['retailer'].unique()
```

```py
array(['Best Buy', 'Amazon', 'Target', 'Walmart'], dtype=object)
```

每个 RDBMS 都有自己的函数集，可以应用于`SELECT`列表中的属性，例如比较运算符、数学函数和运算符，以及字符串函数和运算符。在数据 100 中，我们使用 PostgreSQL，这是一种成熟的 RDBMS，有数百个这样的函数。完整列表可在此处[找到](https://www.postgresql.org/docs/9.2/static/functions.html)。请记住，每个 RDBMS 都有一组不同的函数可用于`SELECT`。

以下代码将所有零售商名称转换为大写，并将产品价格减半。

```py
sql_expr = """
SELECT
    UPPER(retailer) AS retailer_caps,
    product,
    price / 2 AS half_price
FROM prices
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | 零售商上限 | product | 半价 |
| --- | --- | --- | --- |
| 0 | 百思买 | Galaxy S9 | 359.500 美元 |
| --- | --- | --- | --- |
| 1 | BEST BUY | iPod | 10 万 |
| --- | --- | --- | --- |
| 2 | 亚马逊 | iPad | 225.000 美元 |
| --- | --- | --- | --- |
| 3 | AMAZON | Battery pack | 十二点四三五 |
| --- | --- | --- | --- |
| 4 | AMAZON | Chromebook | 124.995 年 |
| --- | --- | --- | --- |
| 5 | 目标 | iPod | 107.500 美元 |
| --- | --- | --- | --- |
| 6 | TARGET | Surface Pro | 399.500 美元 |
| --- | --- | --- | --- |
| 7 | TARGET | Google Pixel 2 | 329.500 美元 |
| --- | --- | --- | --- |
| 8 | 沃尔玛 | Chromebook | 119.395 条 |
| --- | --- | --- | --- |

请注意，我们可以用**别名**将列（指定另一个名称）与`AS`一起使用，以便这些列在输出表中以这个新名称出现。这不会修改源关系中列的名称。

### 式中[¶](#WHERE)

`WHERE`子句允许我们为返回的数据指定某些约束；这些约束通常被称为**谓词**。例如，只检索低于 500 美元的小工具：

```py
sql_expr = """
SELECT *
FROM prices
WHERE price < 500
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer | product | price |
| --- | --- | --- | --- |
| 0 | Best Buy | iPod | 200.00 |
| --- | --- | --- | --- |
| 1 | Amazon | iPad | 450.00 |
| --- | --- | --- | --- |
| 2 | Amazon | Battery pack | 24.87 |
| --- | --- | --- | --- |
| 3 | Amazon | Chromebook | 249.99 |
| --- | --- | --- | --- |
| 4 | Target | iPod | 215.00 |
| --- | --- | --- | --- |
| 5 | Walmart | Chromebook | 238.79 |
| --- | --- | --- | --- |

我们还可以使用操作符`AND`、`OR`和`NOT`进一步约束我们的 SQL 查询。为了在亚马逊上找到一个不带电池包的商品，价格低于 300 美元，我们写下：

```py
sql_expr = """
SELECT *
FROM prices
WHERE retailer = 'Amazon'
    AND NOT product = 'Battery pack'
    AND price < 300
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer | product | price |
| --- | --- | --- | --- |
| 0 | Amazon | Chromebook | 249.99 |
| --- | --- | --- | --- |

`pandas`中的等效操作为：

```py
prices[(prices['retailer'] == 'Amazon') 
   & ~(prices['product'] == 'Battery pack')
   & (prices['price'] <= 300)]
```

|  | retailer | product | price |
| --- | --- | --- | --- |
| 4 | Amazon | Chromebook | 249.99 |
| --- | --- | --- | --- |

有一个细微的区别值得注意：SQL 查询中 Chromebook 的索引是 0，而数据帧中相应的索引是 4。这是因为 SQL 查询总是返回一个索引从 0 开始计数的新表，而`pandas`将数据帧`prices`的一部分子集并返回原始索引。我们可以使用[pd.dataframe.reset_index](https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.reset_index.html)重置`pandas`中的索引。

### 聚合函数[¶](#Aggregate-Functions)

到目前为止，我们只处理表中现有行的数据；也就是说，我们返回的所有表都是表中找到的条目的一个子集。但是为了进行数据分析，我们需要计算数据的聚合值。在 SQL 中，这些函数称为**聚合函数**。

如果我们想找到`prices`关系中所有小工具的平均价格：

```py
sql_expr = """
SELECT AVG(price) AS avg_price
FROM prices
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | 平均价格 |
| --- | --- |
| 0 | 395.072222 个 |
| --- | --- |

相当于，in`pandas`：

```py
prices['price'].mean()
```

```py
395.0722222222222
```

postgresql 聚合函数的完整列表可以在这里找到[。尽管我们在数据 100 中使用 PostgreSQL 作为 SQL 的主要版本，但请记住，SQL 还有许多其他变体（MySQL、SQLite 等），它们使用不同的函数名，并具有不同的可用功能。](https://www.postgresql.org/docs/9.2/static/functions.html)

### 分组依据并具有[¶](#GROUP-BY-and-HAVING)

使用聚合函数，我们可以执行更复杂的 SQL 查询。为了对更细粒度的聚合数据进行操作，我们可以使用以下两个子句：

*   `GROUP BY`获取列列表，并将表分组，如`pandas`中的[pd.dataframe.groupby](https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.groupby.html)函数。
*   `HAVING`在功能上与`WHERE`相似，但仅用于将谓词应用于聚合数据。（请注意，为了使用`HAVING`，它前面必须有一个`GROUP BY`子句。）

**重要**：使用`GROUP BY`时，`SELECT`子句中的所有列都必须在`GROUP BY`子句中列出或应用聚合函数。

我们可以使用这些声明来找到每个零售商的最高价格。

```py
sql_expr = """
SELECT retailer, MAX(price) as max_price
FROM prices
GROUP BY retailer
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer | 最高价格 |
| --- | --- | --- |
| 0 | Amazon | 450.00 |
| --- | --- | --- |
| 1 | Best Buy | 719.00 |
| --- | --- | --- |
| 2 | Target | 799.00 |
| --- | --- | --- |
| 3 | Walmart | 238.79 |
| --- | --- | --- |

比如说，我们有一个品味很高的客户，只想找到售价超过 700 美元的零售商。注意，我们必须使用`HAVING`定义聚合列上的谓词；我们不能使用`WHERE`筛选聚合列。为了计算满足我们需求的零售商列表和附带价格，我们运行：

```py
sql_expr = """
SELECT retailer, MAX(price) as max_price
FROM prices
GROUP BY retailer
HAVING max_price > 700
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer | max_price |
| --- | --- | --- |
| 0 | Best Buy | 七百一十九 |
| --- | --- | --- |
| 1 | Target | 七百九十九 |
| --- | --- | --- |

为了进行比较，我们在`pandas`中重新创建了相同的表：

```py
max_prices = prices.groupby('retailer').max()
max_prices.loc[max_prices['price'] > 700, ['price']]
```

|  | price |
| --- | --- |
| retailer |  |
| --- | --- |
| 百思买 | 719.0 |
| --- | --- |
| 目标 | 799.0 |
| --- | --- |

### 订货人和限额[¶](#ORDER-BY-and-LIMIT)

这些条款允许我们控制数据的显示：

*   `ORDER BY`让我们按照列值的字典顺序显示数据。默认情况下，order by 使用升序（`ASC`），但我们可以使用`DESC`指定降序。
*   `LIMIT`控制显示多少元组。

让我们在`prices`表中显示三个最便宜的项目：

```py
sql_expr = """
SELECT *
FROM prices
ORDER BY price ASC
LIMIT 3
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | retailer | product | price |
| --- | --- | --- | --- |
| 0 | Amazon | Battery pack | 24.87 |
| --- | --- | --- | --- |
| 1 | Best Buy | iPod | 200.00 |
| --- | --- | --- | --- |
| 2 | Target | iPod | 215.00 |
| --- | --- | --- | --- |

注意，我们不必包含`ASC`关键字，因为默认情况下`ORDER BY`按升序返回数据。用于比较，在`pandas`中：

```py
prices.sort_values('price').head(3)
```

|  | retailer | product | price |
| --- | --- | --- | --- |
| 3 | Amazon | Battery pack | 24.87 |
| --- | --- | --- | --- |
| 1 | Best Buy | iPod | 200.00 |
| --- | --- | --- | --- |
| 5 | Target | iPod | 215.00 |
| --- | --- | --- | --- |

（我们再次看到，索引在`pandas`数据帧中不正常。和以前一样，`pandas`在数据帧`prices`上返回一个视图，而 SQL 在每次执行查询时都显示一个新表。）

### 概念性 SQL 评估

SQL 查询中的子句按特定顺序执行。不幸的是，此顺序与在 SQL 查询中写入子句的顺序不同。从最初执行到最后：

1.  `FROM`：一个或多个源表
2.  【HTG0】：申请选择资格（删除行）
3.  【HTG0】：分组和汇总
4.  `HAVING`：消除组
5.  `SELECT`：选择列

**关于`WHERE`对`HAVING`**的说明：由于在应用`GROUP BY`之前处理了`WHERE`子句，因此`WHERE`子句不能使用聚合值。要基于聚合值定义谓词，必须使用`HAVING`子句。

## 摘要[¶](#Summary)

我们介绍了 SQL 语法和使用关系数据库管理系统进行数据分析所需的最重要的 SQL 语句。

## 9.3 SQL 连接

```py
# HIDDEN
# Clear previously defined variables
%reset -f

# Set directory for data loading to work properly
import os
os.chdir(os.path.expanduser('~/notebooks/09'))
```

```py
# HIDDEN
import warnings
# Ignore numpy dtype warnings. These warnings are caused by an interaction
# between numpy and Cython and can be safely ignored.
# Reference: https://stackoverflow.com/a/40846742
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
%matplotlib inline
import ipywidgets as widgets
from ipywidgets import interact, interactive, fixed, interact_manual
import nbinteract as nbi

sns.set()
sns.set_context('talk')
np.set_printoptions(threshold=20, precision=2, suppress=True)
pd.options.display.max_rows = 7
pd.options.display.max_columns = 8
pd.set_option('precision', 2)
# This option stops scientific notation for pandas
# pd.set_option('display.float_format', '{:.2f}'.format)
```

```py
# HIDDEN
# Make names table
sql_expr = """
CREATE TABLE names(
    cat_id INTEGER PRIMARY KEY,
    name TEXT);
"""
result = sqlite_engine.execute(sql_expr)

# Populate names table
sql_expr = """
INSERT INTO names VALUES 
(0, "Apricot"),
(1, "Boots"),
(2, "Cally"),
(4, "Eugene");
"""

result = sqlite_engine.execute(sql_expr)
```

```py
# HIDDEN
# Make colors table
sql_expr = """
CREATE TABLE colors(
    cat_id INTEGER PRIMARY KEY,
    color TEXT);
"""
result = sqlite_engine.execute(sql_expr)

# Populate colors table
sql_expr = """
INSERT INTO colors VALUES 
(0, "orange"),
(1, "black"),
(2, "calico"),
(3, "white");
"""

result = sqlite_engine.execute(sql_expr)
```

```py
# HIDDEN
# Make ages table
sql_expr = """
CREATE TABLE ages(
    cat_id INTEGER PRIMARY KEY,
    age INT);
"""
result = sqlite_engine.execute(sql_expr)

# Populate ages table
sql_expr = """
INSERT INTO ages VALUES 
(0, 4),
(1, 3),
(2, 9),
(4, 20);
"""

result = sqlite_engine.execute(sql_expr)
```

在`pandas`中，我们使用`pd.merge`方法在两个表的列中使用匹配的值连接两个表。例如：

```py
pd.merge(table1, table2, on='common_column')
```

在本节中，我们将介绍 SQL 连接。SQL 连接用于组合关系数据库中的多个表。

假设我们是猫店的老板，有一个我们店里猫的数据库。我们有**两个**不同的表：`names`和`colors`。`names`表包含列`cat_id`、分配给每个 cat 的唯一编号和 cat 的名称`name`。`colors`表包含列`cat_id`和`color`，每只猫的颜色。

请注意，两个表中都有一些缺少的行-`names`表中缺少一行带有`cat_id`3，而`colors`表中缺少一行带有`cat_id`4。

<header style="text-align:center">**names**</header>

| 猫科动物 | 名称 |
| --- | --- |
| 零 | 杏 |
| 1 个 | 靴子 |
| 二 | 凯利 |
| 四 | 尤金 |

<header style="text-align:center">**colors**</header>

| cat_id | 颜色 |
| --- | --- |
| 0 | 橙色 |
| 1 | 黑色 |
| 2 | 印花布 |
| 三 | 白色 |

要计算名为杏的猫的颜色，我们必须在两个表中使用信息。我们可以将 _ 连接到 _ 列上的表，用`name`和`color`创建一个新表。

## 连接[¶](#Joins)

连接通过在表的列中匹配值来组合表。

连接有四种主要类型：内部连接、外部连接、左侧连接和右侧连接。尽管这四个表都是组合表，但每个表对不匹配的值的处理方式不同。

### 内部连接

**定义：**在内部连接中，最终表只包含在**两个**表中具有匹配列的行。

![Inner Join](img/0ce3e484b2dc0eef8d92bf1f4fc4be3d.jpg)

**示例：**我们希望将`names`和`colors`表连接在一起，以匹配每只猫的颜色。因为这两个表都包含一个`cat_id`列，该列是 cat 的唯一标识符，所以我们可以在`cat_id`列上使用内部连接。

**sql:**要在 sql 中编写内部连接，我们修改了`FROM`子句以使用以下语法：

```py
SELECT ...
FROM <TABLE_1>
    INNER JOIN <TABLE_2>
    ON <...>
```

例如：

```py
SELECT *
FROM names AS N
    INNER JOIN colors AS C
    ON N.cat_id = C.cat_id;
```

|  | cat_id | name | cat_id | color |
| --- | --- | --- | --- | --- |
| 零 | 0 | Apricot | 0 | orange |
| --- | --- | --- | --- | --- |
| 1 个 | 1 | Boots | 1 | black |
| --- | --- | --- | --- | --- |
| 二 | 2 | Cally | 2 | calico |
| --- | --- | --- | --- | --- |

您可以验证每个 cat 名称是否与其颜色匹配。请注意，结果表中不存在带有`cat_id`3 和 4 的 cats，因为`colors`表没有带有`cat_id`4 的行，而`names`表没有带有`cat_id`3 的行。在内部连接中，如果一行在另一个表中没有匹配的值，则该行不包括在最终结果中。

假设我们有一个名为`names`的数据帧和一个名为`colors`的数据帧，我们可以通过编写以下内容在`pandas`中执行内部连接：

```py
pd.merge(names, colors, how='inner', on='cat_id')
```

### 全/外接

**定义：**在完全连接（有时称为外部连接）中，**两个表**中的所有记录都包含在连接表中。如果一行在另一个表中没有匹配项，则用`NULL`填充缺少的值。

![Full outer join](img/831ee85b962ad0678a02c0098bf76e5f.jpg)

**示例：**和前面一样，我们将`names`和`colors`表连接在一起，以匹配每只猫的颜色。这一次，我们希望将所有行保留在两个表中，即使没有匹配的行。

**sql:**要在 sql 中编写外部连接，我们修改了`FROM`子句，以使用以下语法：

```py
SELECT ...
FROM <TABLE_1>
    FULL JOIN <TABLE_2>
    ON <...>
```

For example:

```py
SELECT name, color
FROM names N
    FULL JOIN colors C
    ON N.cat_id = C.cat_id;
```

| cat_id | name | color |
| --- | --- | --- |
| 0 | Apricot | orange |
| 1 | Boots | black |
| 2 | Cally | calico |
| 3 | 无效的 | white |
| 4 | Eugene | NULL |

请注意，最终输出包含带有`cat_id`3 和 4 的条目。如果一行没有匹配项，它仍然包含在最终输出中，并且任何缺少的值都用`NULL`填充。

在`pandas`中：

```py
pd.merge(names, colors, how='outer', on='cat_id')
```

### 左连接[¶](#Left-Join)

**定义：**在左连接中，来自**左表**的所有记录都包含在连接表中。如果行在右表中没有匹配项，则缺少的值将用`NULL`填充。

![left join](img/7221e81b49ef94eaaca9721cff85eab7.jpg)

**示例：**和前面一样，我们将`names`和`colors`表连接在一起，以匹配每只猫的颜色。这次，我们要保留所有的猫名，即使一只猫没有匹配的颜色。

**sql:**要在 sql 中编写左连接，我们修改了`FROM`子句以使用以下语法：

```py
SELECT ...
FROM <TABLE_1>
    LEFT JOIN <TABLE_2>
    ON <...>
```

For example:

```py
SELECT name, color
FROM names N
    LEFT JOIN colors C
    ON N.cat_id = C.cat_id;
```

| cat_id | name | color |
| --- | --- | --- |
| 0 | Apricot | orange |
| 1 | Boots | black |
| 2 | Cally | calico |
| 4 | Eugene | NULL |

请注意，最终输出包括所有四个 cat 名称。`names`关系中的三个`cat_id`s 在`colors`表中与`cat_id`s 匹配，一个不匹配（eugene）。没有匹配颜色的猫名的颜色为`NULL`。

In `pandas`:

```py
pd.merge(names, colors, how='left', on='cat_id')
```

### 右连接[¶](#Right-Join)

**定义：**在右连接中，来自**右表**的所有记录都包含在连接表中。如果左表中的行不匹配，则用`NULL`填充缺少的值。

![right join](img/2616ce35cb39df1e77481c08cf41bdba.jpg)

**示例：**和前面一样，我们将`names`和`colors`表连接在一起，以匹配每只猫的颜色。这一次，我们要保留所有的猫的颜色，即使一只猫没有匹配的名字。

**sql:**要在 sql 中编写正确的 join，我们修改了`FROM`子句以使用以下语法：

```py
SELECT ...
FROM <TABLE_1>
    RIGHT JOIN <TABLE_2>
    ON <...>
```

For example:

```py
SELECT name, color
FROM names N
    RIGHT JOIN colors C
    ON N.cat_id = C.cat_id;
```

| cat_id | name | color |
| --- | --- | --- |
| 0 | Apricot | orange |
| 1 | Boots | black |
| 2 | Cally | calico |
| 3 | NULL | white |

这一次，观察最终输出包括所有四种 CAT 颜色。`colors`关系中的三个`cat_id`s 与`names`表中的`cat_id`s 匹配，一个不匹配（白色）。没有匹配名称的 cat 颜色的名称为`NULL`。

您还可能注意到，右连接产生的结果与交换表顺序的左连接相同。即，`names`左接`colors`与`colors`右接`names`相同。因此，一些 SQL 引擎（如 sqlite）不支持右连接。

In `pandas`:

```py
pd.merge(names, colors, how='right', on='cat_id')
```

### 隐式内部连接[¶](#Implicit-Inner-Joins)

在 SQL 中通常有多种方法来完成同一个任务，就像在 Python 中有多种方法来完成同一个任务一样。我们指出了另一种编写内部连接的方法，这种方法在实践中出现，称为 _ 隐式连接 _。回想一下，我们之前编写了以下内容来进行内部连接：

```py
SELECT *
FROM names AS N
    INNER JOIN colors AS C
    ON N.cat_id = C.cat_id;
```

隐式内部连接的语法稍有不同。请特别注意，`FROM`子句使用逗号从两个表中进行选择，并且查询包含一个`WHERE`子句来指定连接条件。

```py
SELECT *
FROM names AS N, colors AS C
WHERE N.cat_id = C.cat_id;
```

当在`FROM`子句中指定多个表时，SQL 将创建一个表，其中包含每个表中的每一行组合。例如：

```py
sql_expr = """
SELECT *
FROM names N, colors C
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | cat_id | name | cat_id | color |
| --- | --- | --- | --- | --- |
| 0 | 0 | Apricot | 0 | orange |
| --- | --- | --- | --- | --- |
| 1 | 0 | Apricot | 1 | black |
| --- | --- | --- | --- | --- |
| 2 | 0 | Apricot | 2 | calico |
| --- | --- | --- | --- | --- |
| 三 | 0 | Apricot | 3 | white |
| --- | --- | --- | --- | --- |
| 四 | 1 | Boots | 0 | orange |
| --- | --- | --- | --- | --- |
| 5 个 | 1 | Boots | 1 | black |
| --- | --- | --- | --- | --- |
| 六 | 1 | Boots | 2 | calico |
| --- | --- | --- | --- | --- |
| 七 | 1 | Boots | 3 | white |
| --- | --- | --- | --- | --- |
| 8 个 | 2 | Cally | 0 | orange |
| --- | --- | --- | --- | --- |
| 九 | 2 | Cally | 1 | black |
| --- | --- | --- | --- | --- |
| 10 个 | 2 | Cally | 2 | calico |
| --- | --- | --- | --- | --- |
| 11 个 | 2 | Cally | 3 | white |
| --- | --- | --- | --- | --- |
| 12 个 | 4 | Eugene | 0 | orange |
| --- | --- | --- | --- | --- |
| 十三 | 4 | Eugene | 1 | black |
| --- | --- | --- | --- | --- |
| 十四 | 4 | Eugene | 2 | calico |
| --- | --- | --- | --- | --- |
| 15 个 | 4 | Eugene | 3 | white |
| --- | --- | --- | --- | --- |

此操作通常称为 _ 笛卡尔积 _：第一个表中的每一行都与第二个表中的每一行成对出现。请注意，许多行包含的 cat 颜色与它们的名称不匹配。隐式连接中的附加`WHERE`子句筛选出没有匹配`cat_id`值的行。

```py
SELECT *
FROM names AS N, colors AS C
WHERE N.cat_id = C.cat_id;
```

|  | cat_id | name | cat_id | color |
| --- | --- | --- | --- | --- |
| 0 | 0 | Apricot | 0 | orange |
| --- | --- | --- | --- | --- |
| 1 | 1 | Boots | 1 | black |
| --- | --- | --- | --- | --- |
| 2 | 2 | Cally | 2 | calico |
| --- | --- | --- | --- | --- |

## 连接多个表[¶](#Joining-Multiple-Tables)

若要连接多个表，请使用附加的`JOIN`运算符扩展`FROM`子句。例如，下表`ages`包括每只猫的年龄数据。

| cat_id | 年龄 |
| --- | --- |
| 0 | 4 |
| 1 | 3 |
| 2 | 九 |
| 4 | 20 个 |

要在`names`、`colors`和`ages`表上执行内部连接，我们编写：

```py
# Joining three tables

sql_expr = """
SELECT name, color, age
    FROM names n
    INNER JOIN colors c ON n.cat_id = c.cat_id
    INNER JOIN ages a ON n.cat_id = a.cat_id;
"""
pd.read_sql(sql_expr, sqlite_engine)
```

|  | name | color | age |
| --- | --- | --- | --- |
| 0 | Apricot | orange | 4 |
| --- | --- | --- | --- |
| 1 | Boots | black | 3 |
| --- | --- | --- | --- |
| 2 | Cally | calico | 9 |
| --- | --- | --- | --- |

## 摘要[¶](#Summary)

我们已经介绍了四种主要的 SQL 连接类型：内部连接、完整连接、左连接和右连接。我们使用所有四个连接将信息组合在单独的关系中，并且每个连接只在处理输入表中不匹配行的方式上有所不同。