# 零、前言

本书是有关使用 Python 进行时间序列分析的简介。 我们旨在为您提供该学科基本概念的清晰概述，并描述适用于该行业中常见的分析用例的有用技术。 由于有太多的项目需要根据过去的数据进行趋势分析和预测，因此时间序列分析是任何现代数据科学家的知识宝库中的重要工具。 本书将为您提供工具和技术，使您可以自信地思考问题并提出时间序列预测中的解决方案。

为什么选择 Python？ Python 正在迅速成为跨不同行业的数据科学项目的首选。 大多数最先进的机器学习和深度学习库都具有 Python API。 结果，许多数据科学家更喜欢使用 Python 来实现整个项目管道，该管道包括数据整理，模型构建和模型验证。 此外，Python 提供了易于使用的 API 来处理，建模和可视化时间序列数据。 此外，Python 是开发 Web 应用后端的一种流行语言，因此吸引了广大软件专业人员。

现在，让我们看看您可以从本书的每一章中学到什么。

# 本书涵盖的内容

第 1 章和“时间序列简介”首先讨论三种不同类型的数据集-横截面，时间序列和面板。 讨论了从横截面到时间序列的过渡以及数据分析增加的复杂性。 描述了使时间序列数据变得特别的特殊数学属性。 几个示例演示了如何使用探索性数据分析来可视化这些属性。

第 2 章，“了解时间序列数据”涉及三个主题，即通过*重采样*，*分组依据*对时间序列数据进行高级预处理和可视化 ]和移动平均线的计算； 平稳性和统计假设检验，以检测时间序列中的平稳性； 以及用于平稳化非平稳时间序列的各种时间序列分解方法。

第 3 章，“基于指数平滑的方法”涵盖了使用 Holt-Winters 方法进行一阶捕获级别，二阶平滑级别和趋势以及高阶平滑的基于平滑的模型。 说明了该图，该图捕获了时间序列数据集中的水平，趋势和季节性。

第 4 章和“自回归模型”讨论了用于预测的自回归模型。 本章涵盖移动平均值（MA），自回归（AR），自回归移动平均值（ARMA）和自回归综合移动平均值（ARIMA）的详细实现，以捕获预测期间时间序列数据中不同级别的干扰。

第 5 章，“用于时间序列预测的深度学习”讨论了可以直接用于开发时间序列数据的预测模型的最新深度学习算法。 递归神经网络（RNN）是对数据序列进行建模的自然选择。 在本章中，将介绍不同的 RNN，例如 Vanilla RNN，门控循环单元和长期短期记忆单元，以开发时间序列数据的预测模型。 在概念上讨论了开发这些 RNN 的数学公式。 案例研究使用 Python 的“ keras”深度学习库解决。

[附录](6.html)和 *Python 入门*，您将快速，轻松地了解 Python。 如果您不熟悉 Python 或正在寻找如何开始使用编程语言，阅读本附录将帮助您克服最初的障碍。

# 这本书需要什么

您将需要 Anaconda Python 发行版来运行本书中的示例，并编写自己的 Python 程序以进行时间序列分析。 可从[这个页面](https://www.continuum.io/downloads)免费下载。

本书的代码示例是使用 Jupyter Notebook 开发环境编写的。 要运行 Jupyter Notebook，您需要安装 Anaconda Python Distribution，它具有 Python 语言要点，解释器，用于开发示例的包以及 Jupyter Notebook 服务器。

# 这本书是给谁的

本书的主题预期对以下人员有用：

*   数据科学家，具有统计，机器学习以及模型构建和验证背景的专业人员
*   数据工程师，具有软件开发背景的专业人员
*   希望在生成数据驱动的业务见解方面发展专业知识的软件专业人员

# 约定

在本书中，您将找到许多可以区分不同类型信息的文本样式。 以下是这些样式的一些示例，并对其含义进行了解释。

代码块设置如下：

```py
import os
import pandas as pd
%matplotlib inline
from matplotlib import pyplot as plt
import seaborn as sns
```

文本代码以字体和颜色突出显示，如下所示：`pandas.DataFrame`。 文件和文件夹的名称也以相同的样式显示，例如`Chapter_1_Models_for_Time_Series_Analysis.ipynb or datasets/DJIA_Jan2016_Dec2016.xlsx`

在本书的多个地方，我们都引用了外部 URL 来引用数据集或其他信息的来源。 [URL 将以以下文本样式显示](http://finance.yahoo.com)

**新术语**和**重要词**以粗体显示。 您在屏幕上看到的字词，例如在菜单或对话框中，将以如下形式显示在文本中：“为了下载新模块，我们将转到“文件” |“设置” |“项目名称” |“项目解释器”。”

警告或重要提示如下所示。

提示和技巧如下所示。

# 读者反馈

始终欢迎读者的反馈。 让我们知道您对这本书的看法-您喜欢或不喜欢的书。 读者反馈对我们很重要，因为它可以帮助我们开发出您真正能充分利用的标题。 要向我们发送一般性反馈，只需通过电子邮件发送`feedback@packtpub.com`，然后在您的消息主题中提及该书的标题。 如果您有专业知识的主题，并且对写作或撰写书籍感兴趣，请参阅[这个页面](http://www.packtpub.com/authors)上的作者指南。

# 客户支持

既然您是 Packt 书的骄傲拥有者，我们可以通过很多方法来帮助您从购买中获得最大收益。

# 下载示例代码

您可以从[这个页面](http://www.packtpub.com)的帐户中下载本书的示例代码文件。 如果您在其他地方购买了此书，则可以访问[这个页面](http://www.packtpub.com/support)并注册以将文件直接通过电子邮件发送给您。 您可以按照以下步骤下载代码文件：

1.  使用您的电子邮件地址和密码登录或注册到我们的网站。
2.  将鼠标指针悬停在顶部的“支持”选项卡上。
3.  单击“代码下载和勘误”。
4.  在搜索框中输入书籍的名称。
5.  选择您要下载其代码文件的书。
6.  从购买本书的下拉菜单中选择。
7.  单击代码下载。

下载文件后，请确保使用以下最新版本解压缩或解压缩文件夹：

*   Windows 的 WinRAR / 7-Zip
*   适用于 Mac 的 Zipeg / iZip / UnRarX
*   适用于 Linux 的 7-Zip / PeaZip

本书的代码包也托管在 [GitHub](https://github.com/PacktPublishing/Practical-Time-Series-Analysis) 上。 我们还从[这个页面](https://github.com/PacktPublishing/)提供了丰富的书籍和视频目录中的其他代码包。 去看一下！

# 勘误

尽管我们已尽一切努力确保内容的准确性，但还是会发生错误。 如果您发现我们的其中一本书中有错误-可能是文字或代码中的错误-请向我们报告，我们将不胜感激。 这样，您可以使其他读者免于沮丧，并帮助我们改进本书的后续版本。 如果您发现任何勘误，请访问[这个页面](http://www.packtpub.com/submit-errata)进行报告，选择您的图书，点击“勘误提交表格”链接，然后输入勘误的详细信息 。 勘误一经确认，您的提交将被接受，勘误将被上载到我们的网站或添加到该标题的勘误部分下的任何现有勘误列表中。 要查看以前提交的勘误，请转到[这个页面](https://www.packtpub.com/books/content/support)，然后在搜索字段中输入书籍的名称。 所需信息将出现在“勘误”部分下。

# 海盗行为

互联网上版权材料的盗版在所有媒体中都是一个持续存在的问题。 在 Packt，我们非常重视版权和许可的保护。 如果您在互联网上以任何形式发现我们的作品的任何非法副本，请立即向我们提供位置地址或网站名称，以便我们寻求补救。 请通过`copyright@packtpub.com`与我们联系，并提供指向可疑盗版材料的链接。 感谢您在保护我们的作者方面的帮助以及我们为您带来有价值的内容的能力。

# 问题

如果您对本书的任何方面都有疑问，可以通过`questions@packtpub.com`与我们联系，我们将尽最大努力解决该问题。